(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('component-registry')) :
  typeof define === 'function' && define.amd ? define(['exports', 'component-registry'], factory) :
  (factory((global.IsomorphicSchema = {}),global.componentRegistry));
}(this, (function (exports,componentRegistry) { 'use strict';

  var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
    return typeof obj;
  } : function (obj) {
    return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
  };

  var classCallCheck = function (instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  };

  var createClass = function () {
    function defineProperties(target, props) {
      for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) descriptor.writable = true;
        Object.defineProperty(target, descriptor.key, descriptor);
      }
    }

    return function (Constructor, protoProps, staticProps) {
      if (protoProps) defineProperties(Constructor.prototype, protoProps);
      if (staticProps) defineProperties(Constructor, staticProps);
      return Constructor;
    };
  }();

  /*
      Example of field definitions:
      
      fieldDefs = {
          field_name: {
              validator: function (inp) {
                  return {
                      message: "This field is required"
                  }
              },
              required: false
          }
      }

      invariantDefs = [
          function (data, selectedFields) {
              return {
                  message: "bla bla",
                  fields: ['field_name', 'field_name']
              }
          }
      ]
  */

  var Schema = function () {
      /**
       * Create a form or object schema. You would normally use one of the predefined field types
       * in isomorphic-schema or a subclass of these in order to support auto-generated forms.
       * 
       * However, this isn't required as long as you provide an object of the kind:
       * { validator(), fromString() }
       * 
       * Passing schema name as first param is legacy but common. It won't be deprecated, but doesn't
       * support 'extends'.
       * 
       * @param {(string|Object)} opts – configuration object
       * @param {string} opts.schemaName – name of schema
       * @param {Object} opts.fields – schema fields
       * @param {Object[]} opts.extends – list of schemas to extend for composition (experimental)
       * @param {Object=} fieldDefs – schema fields (overridden if fields are passed in first param)
       * @returns {Object}
       */
      function Schema(opts, fieldDefs) {
          classCallCheck(this, Schema);

          if (typeof opts === 'string') {
              // Backward compatibility
              var schemaName = opts;
          } else {
              var schemaName = opts.schemaName;
              fieldDefs = fieldDefs || opts.fields;
          }

          // Check that we paased a schema name (used for debugging)
          if (typeof schemaName !== 'string') {
              throw {
                  message: "You didn't pass a name to the schema",
                  data: {
                      schemaName: schemaName,
                      fieldDefs: fieldDefs
                  }
              };
          }

          this._name = schemaName;
          this._fields = {};
          this._invariants = [];
          this._validationConstraints = [];

          // Check if we are extending other schemas
          if (Array.isArray(opts.extends)) {
              // Fields in schemas to the right will override those to the left
              opts.extends.map(function (schema) {
                  for (var key in schema._fields) {
                      this._fields[key] = schema._fields[key];
                  }
                  for (var key in schema._invariants) {
                      this._invariants.push(schema._invariants[key]);
                  }
                  for (var key in schema._validationConstraints) {
                      this._validationConstraints.push(schema._validationConstraints[key]);
                  }
              }.bind(this));
          }

          for (var key in fieldDefs) {
              this._fields[key] = fieldDefs[key];
          }
      }

      /**
       * If this function returns undefined then there is no invariant error. If there is an invariant error
       *
       * @callback invariantCondition
       * @param {Object} data – the complete data of the current object.
       * @param {string[]} selectedFields – list of all fields that are active (respecting selected and omitted fields)
       */

      /**
       * Add an invariant validation rule. You use it to make sure a max value is larger than a min value or that
       * password equals to confirmPassword etc.
       * @param {invariantCondition} invariant
       */


      createClass(Schema, [{
          key: 'addInvariant',
          value: function addInvariant(invariant) {
              // TODO: Check that it is valid
              this._invariants.push(invariant);
          }

          /**
           * If this function returns true the field is included. Perhaps you only want to include a field
           * if another field is of a certain value.
           *
           * @callback validationConstraint
           * @param {Object} data – the complete data of the current object.
           * @param {string} fieldKey – dot notation path to field being evaluated.
           * @returns {boolean}
           */

          /**
           * Add a condition under which the field should be active. Used when generating forms and during validation
           * 
           * @param {validationConstraint} constraint – the validation constraint
           */

      }, {
          key: 'addValidationConstraint',
          value: function addValidationConstraint(constraint) {
              // TODO: Check that it is valid
              this._validationConstraints.push(constraint);
          }
      }, {
          key: '_addFieldError',
          value: function _addFieldError(formErrors, propName, error) {
              if (typeof formErrors.fieldErrors === "undefined") {
                  formErrors.fieldErrors = {};
              }
              formErrors.fieldErrors[propName] = error;
          }
      }, {
          key: '_addInvariantError',
          value: function _addInvariantError(formErrors, error) {
              if (typeof formErrors.invariantErrors === "undefined") {
                  formErrors.invariantErrors = [];
              }
              formErrors.invariantErrors.push(error);
          }
      }, {
          key: '_getSelectedFieldsList',
          value: function _getSelectedFieldsList(selected, omitted, objectPath) {
              var namespace = Array.isArray(objectPath) && objectPath.join('.') || '';
              var newFieldOptions = {};
              if (typeof selected === "string") {
                  selected = selected.split(',');
              }

              if (typeof omitted === "string") {
                  omitted = omitted.split(',');
              }

              var selectedFields = [];

              for (var key in this._fields) {
                  var dottedName = namespace ? namespace + '.' + key : key;

                  if (selected === undefined || selected.indexOf(dottedName) >= 0) {
                      // This field has been selected
                      if (!omitted || omitted.indexOf(dottedName) < 0) {
                          // And not omitted
                          selectedFields.push(key);
                      }
                  }
              }

              // Remove root level props so they aren't counted when we pass on to next level
              // so we can select all if none are passed
              if (Array.isArray(selected)) {
                  selected = selected.filter(function (key) {
                      return key.indexOf('.') >= 0;
                  });
              }

              return {
                  selectedFields: selectedFields,
                  newFieldOptions: newFieldOptions
              };
          }

          /**
           * Validate data async using defined schema. Allways returns a promise
           * @param {Object} data – the data input to be validated
           * @param {Object} options – options object
           * @param {(string|string[])} options.selected – only include selected fields (comma separated list, or array, of prop names)
           * @param {(string|string[])} options.omitted – omit omitted fields (comma separated list, or array, of prop names)
           * @param {string[]} options.objectPath – path to input object, used when transforming nested object
           * @param {Object} context – the entire data object which can be used by a field for conditional validation against other props
           * @returns {Promise}
           */

      }, {
          key: 'validateAsync',
          value: function validateAsync(data, options, context) {
              return this.validate(data, options, context || data, true);
          }

          /**
           * Validate data using defined schema.
           * @param {Object} data – the data input to be validated
           * @param {Object} options – options object
           * @param {(string|string[])} options.selected – only include selected fields (comma separated list, or array, of prop names)
           * @param {(string|string[])} options.omitted – omit omitted fields (comma separated list, or array, of prop names)
           * @param {string[]} options.objectPath – path to input object, used when transforming nested object
           * @param {Object} context – the entire data object which can be used by a field for conditional validation against other props
           * @param {boolean} async – (use validateAsync instead) data is validated async because schema contains async fields. This will return a promise
           * @returns {(undefined|Object|Promise)} – returns error object if validation errors are found
           */

      }, {
          key: 'validate',
          value: function validate(data, options, context, async) {
              var skipInvariants = false,
                  selected = undefined,
                  omitted = undefined;

              if (options) {
                  skipInvariants = options.skipInvariants || false;
                  selected = options.selectedFields || options.selectFields;
                  omitted = options.omittedFields || options.omitFields;
              }

              // We are making the validation stateless to avoid contaminating schemas
              // that are now used as a way to define the object model
              var formErrors = {
                  fieldErrors: undefined,
                  invariantErrors: undefined
              };

              var _getSelectedFieldsLis = this._getSelectedFieldsList(selected, omitted, options && options['objectPath']),
                  selectedFields = _getSelectedFieldsLis.selectedFields,
                  newFieldOptions = _getSelectedFieldsLis.newFieldOptions;

              var validationPromises = [];
              // Validate selected fields
              for (var i in selectedFields) {
                  var key = selectedFields[i];

                  // Make sure we have a new options object...
                  if (!options) {
                      var newOptions = {};
                  } else {
                      var newOptions = Object.assign({}, options, newFieldOptions);
                  }

                  // ...so we can add objectPath to determine where we are
                  if (Array.isArray(newOptions.objectPath)) {
                      newOptions.objectPath = newOptions.objectPath.concat([key]);
                  } else {
                      newOptions.objectPath = [key];
                  }

                  // Remove the selectedFields if passed because we don't want them to be sent to next validate function
                  if (newOptions.selectedFields) {
                      delete newOptions.selectedFields;
                  }

                  // console.log("[Schema] validating: " + newOptions.objectPath.join("."))

                  try {

                      // Check validation constraints
                      var failedValidators = this._validationConstraints.filter(function (constraint) {
                          return !constraint(data, key, options);
                      });
                      // Only validate if we pass all the validation constraints
                      if (failedValidators.length === 0) {
                          if (!this._fields[key].readOnly) {
                              if (!async) {
                                  var tmpError = this._fields[key].validate(data[key], newOptions, context || data);
                                  tmpError && this._addFieldError(formErrors, key, tmpError);
                              } else {
                                  var tmpPromise = this._fields[key].validateAsync(data[key], newOptions, context || data);
                                  if (!(tmpPromise && tmpPromise.then)) {
                                      // We didn't get a promise so this validation was synchronous
                                      var tmpError = tmpPromise;
                                      tmpError && this._addFieldError(formErrors, key, tmpError);
                                      validationPromises.push(Promise.resolve(formErrors));
                                      // Errors are added when promises are resolved at end of this function
                                  } else {
                                      var _this = this;
                                      (function (formErrors, key) {
                                          validationPromises.push(tmpPromise.then(function (tmpError) {
                                              tmpError && _this._addFieldError(formErrors, key, tmpError);
                                              return Promise.resolve(formErrors);
                                          }));
                                      })(formErrors, key);
                                  }
                              }
                          }
                      }
                  } catch (e) {
                      console.error('[Schema] Field validation error for: ' + key);
                      console.log(data && data[key]);
                      throw e;
                  }
              }

              if (!skipInvariants) {

                  // Validate invariants
                  for (var i = 0, imax = this._invariants.length; i < imax; i++) {
                      try {
                          if (!async) {
                              var tmpError = this._invariants[i](data, selectedFields, options);
                              tmpError && this._addInvariantError(formErrors, tmpError);
                          } else {
                              var tmpPromise = this._invariants[i](data, selectedFields, options);
                              if (!(tmpPromise && tmpPromise.then)) {
                                  // We didn't get a promise so this validation was synchronous
                                  var tmpError = tmpPromise;
                                  tmpError && this._addInvariantError(formErrors, tmpError);
                                  validationPromises.push(Promise.resolve(formErrors));
                              } else {
                                  var _this = this;
                                  validationPromises.push(tmpPromise.then(function (tmpError) {
                                      tmpError && _this._addInvariantError(formErrors, tmpError);
                                      return Promise.resolve(formErrors);
                                  }));
                              }
                          }
                      } catch (e) {
                          console.error('[Schema] Invariant validation error for: ' + key);
                          console.log(data[key]);
                          throw e;
                      }
                  }
              }

              if (!async) {
                  // Return error object if any found, else return undefined
                  if (typeof formErrors.fieldErrors === "undefined" && typeof formErrors.invariantErrors === "undefined") {
                      return undefined;
                  } else {
                      return formErrors;
                  }
              } else {
                  // Return a promise with the formErrors object
                  return Promise.all(validationPromises).then(function (results) {
                      // TODO: Right now we are basically return multiple promises with the same result so we only need to
                      // look at the first item. Should probably clean this up so we only return a single promise, but for now...
                      var formErrors = results[0];
                      if (formErrors === undefined || formErrors.fieldErrors === undefined && formErrors.invariantErrors === undefined) {
                          return Promise.resolve(undefined);
                      } else {
                          return Promise.resolve(formErrors);
                      }
                  });
              }
          }

          /**
           * Adds props as defined in schema to passed object in place. Uses Object.defineProperty.
           * @param {Object} obj – the object to decorate
           */

      }, {
          key: 'addProperties',
          value: function addProperties(obj) {
              var schema = this;

              var fields = schema && schema._fields || [];
              for (var key in fields) {
                  var field = fields[key];
                  Object.defineProperty(obj, key, {
                      configurable: true, // We might want to remove properties when passing data through API
                      enumerable: true,
                      // Changed so props are allways writable, otherwise you can have problems creating the object
                      // programmtically. Thus field.readOnly is only a form related property
                      // writable: !field.readOnly
                      writable: true
                  });
              }
          }

          /**
           * Returns fields defined for this schema
           * @returns {Object} object containing fields
           */

      }, {
          key: 'getFields',
          value: function getFields() {
              return this._fields;
          }

          /**
           * Recursively transforms form data. Normally used to convert input data from a browser form. Calls
           * .fromString on each field in schema in order to allow data to be cleaned and typed as specified
           * for each field type.
           * @param {Object} data – input object of any type, usually form data from browser
           * @param {Object} options – options object
           * @param {(string|string[])} options.selected – only include selected fields (comma separated list, or array, of prop names)
           * @param {(string|string[])} options.omitted – omit omitted fields (comma separated list, or array, of prop names)
           * @param {string[]} options.objectPath – path to input object, used when transforming nested object
           * @param {boolean} options.doNotRemoveReadOnly – standard behaviour is to remove readOnly props. This overrides
           * @returns {Object} returns the transformed object
           */

      }, {
          key: 'transform',
          value: function transform(data) {
              var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
              var selected = options.selectedFields,
                  omitted = options.omittedFields,
                  _options$objectPath = options.objectPath,
                  objectPath = _options$objectPath === undefined ? [] : _options$objectPath,
                  doNotRemoveReadOnly = options.doNotRemoveReadOnly;

              var _getSelectedFieldsLis2 = this._getSelectedFieldsList(selected, omitted, objectPath),
                  selectedFields = _getSelectedFieldsLis2.selectedFields,
                  newFieldOptions = _getSelectedFieldsLis2.newFieldOptions;

              var outp = {};
              for (var i in selectedFields) {
                  var key = selectedFields[i];
                  newFieldOptions['objectPath'] = objectPath.concat(key);
                  try {
                      if (!this._fields[key].readOnly || doNotRemoveReadOnly) {
                          outp[key] = this._fields[key].fromString(data[key], Object.assign(options, newFieldOptions));
                      }
                  } catch (err) {
                      // TODO: Error handling
                      console.warn(err);
                  }
              }
              return outp;
          }
      }]);
      return Schema;
  }();

  function cloneShallow(obj) {
      var fn = function fn() {};
      fn.prototype = obj.prototype;
      var outp = new fn();

      Object.keys(obj).forEach(function (key) {
          if (!obj.hasOwnProperty(key)) return;
          outp[key] = obj[key];
      });
      return outp;
  }

  /**
   * A helper method that returns the first parameter as is. The purpose of this
   * method is to allow parsing of code to find translation keys and also provide
   * a help text to the translator.
   * 
   * You would use this when you define text strings in your schemas.
   * @param {string} i18nKey -- the translation string key
   * @param {string} helpText -- a text describing the key
   *
   * @return {string} the passed i18nKey
   *
   * @example
   * i18n('form_label_title', 'The Title')
   */
  function i18n(i18nKey, helpText) {
      return i18nKey;
  }

  var regex = /\${([^}]+)}/g;
  /**
   * A helper method that substitutes placeholders of the form ${propertyName}
   * with the value found in the field definition. Such as minValue on IntegerField
   * NOTE: It will add a preceding underscore automatically if needed!
   * @param {string} str -- the input string
   * @param {object} fieldDef -- the field validator
   *
   * @return {string} the rendered string
   *
   * @example
   * renderString('Min ${minValue}', personSchema.age)
   * 
   * Where personSchema.age is an IntegerField validator
   */
  function renderString(str, fieldDef) {
      var m;
      while ((m = regex.exec(str)) !== null) {
          // This is necessary to avoid infinite loops with zero-width matches
          if (m.index === regex.lastIndex) {
              regex.lastIndex++;
          }

          // The result can be accessed through the `m`-variable.
          m.forEach(function (match, groupIndex) {
              if (fieldDef[match] || fieldDef['_' + match]) {
                  str = str.replace('${' + match + '}', fieldDef[match] || fieldDef['_' + match]);
              }
          });
      }
      return str;
  }

  var Interface = componentRegistry.createInterfaceClass('isomorphic-schema');

  function addMethods(base) {
      base.prototype.validate = function () {};
      base.prototype.fromString = function () {};
      base.prototype.toFormattedString = function () {};
  }

  var IBaseField = new Interface({
      name: 'IBaseField'
  });
  addMethods(IBaseField);

  var ITextField = new Interface({
      name: 'ITextField'
  });
  addMethods(ITextField);

  var ITextAreaField = new Interface({
      name: 'ITextAreaField'
  });
  addMethods(ITextAreaField);

  var IIntegerField = new Interface({
      name: 'IIntegerField'
  });
  addMethods(IIntegerField);

  var IDecimalField = new Interface({
      name: 'IDecimalField'
  });
  addMethods(IDecimalField);

  var IBoolField = new Interface({
      name: 'IBoolField'
  });
  addMethods(IBoolField);

  var ICreditCardField = new Interface({
      name: 'ICreditCardField'
  });
  addMethods(ICreditCardField);

  var IDateField = new Interface({
      name: 'IDateField'
  });
  addMethods(IDateField);

  var IDateTimeField = new Interface({
      name: 'IDateTimeField'
  });
  addMethods(IDateTimeField);

  var IEmailField = new Interface({
      name: 'IEmailField'
  });
  addMethods(IEmailField);

  var ITelephoneField = new Interface({
      name: 'ITelephoneField'
  });
  addMethods(ITelephoneField);

  var IListField = new Interface({
      name: 'IListField'
  });
  addMethods(IListField);

  var IObjectField = new Interface({
      name: 'IObjectField'
  });
  addMethods(IObjectField);

  var IObjectRelationField = new Interface({
      name: 'IObjectRelationField'
  });
  addMethods(IObjectRelationField);

  var IOrgNrField = new Interface({
      name: 'IOrgNrField'
  });
  addMethods(IOrgNrField);

  var IPasswordField = new Interface({
      name: 'IPasswordField'
  });
  addMethods(IPasswordField);

  var ISelectField = new Interface({
      name: 'ISelectField'
  });
  addMethods(ISelectField);

  var IDynamicSelectBaseField = new Interface({
      name: 'IDynamicSelectBaseField'
  });
  addMethods(IDynamicSelectBaseField);

  var IDynamicSelectAsyncBaseField = new Interface({
      name: 'IDynamicSelectAsyncBaseField'
  });
  addMethods(IDynamicSelectAsyncBaseField);

  var IMultiSelectField = new Interface({
      name: 'IMultiSelectField'
  });
  addMethods(IMultiSelectField);

  var IDynamicMultiSelectField = new Interface({
      name: 'IDynamicMultiSelectField'
  });
  addMethods(IDynamicMultiSelectField);

  /* Richtext HTML field */
  var IHTMLAreaField = new Interface({
      name: 'IHTMLAreaField'
  });
  addMethods(IHTMLAreaField);

  var IAnyOf = new Interface({
      name: 'IAnyOf'
  });
  addMethods(IAnyOf);

  var interfaces = /*#__PURE__*/Object.freeze({
    IBaseField: IBaseField,
    ITextField: ITextField,
    ITextAreaField: ITextAreaField,
    IIntegerField: IIntegerField,
    IDecimalField: IDecimalField,
    IBoolField: IBoolField,
    ICreditCardField: ICreditCardField,
    IDateField: IDateField,
    IDateTimeField: IDateTimeField,
    IEmailField: IEmailField,
    ITelephoneField: ITelephoneField,
    IListField: IListField,
    IObjectField: IObjectField,
    IObjectRelationField: IObjectRelationField,
    IOrgNrField: IOrgNrField,
    IPasswordField: IPasswordField,
    ISelectField: ISelectField,
    IDynamicSelectBaseField: IDynamicSelectBaseField,
    IDynamicSelectAsyncBaseField: IDynamicSelectAsyncBaseField,
    IMultiSelectField: IMultiSelectField,
    IDynamicMultiSelectField: IDynamicMultiSelectField,
    IHTMLAreaField: IHTMLAreaField,
    IAnyOf: IAnyOf
  });

  var BaseField = componentRegistry.createObjectPrototype({
      implements: [IBaseField],

      constructor: function constructor(options) {
          if (options) {
              this._isRequired = options.required;
              delete options.required;
          }
      },

      validateAsync: function validateAsync(inp, options, context) {
          var tmp = this.validate(inp, options, context, true);
          if (tmp && tmp.then) {
              return tmp;
          } else {
              return Promise.resolve(tmp);
          }
      },

      validate: function validate(inp, options, context, async) {
          context = context || inp;
          if (this._isRequired && (typeof inp === "undefined" || inp === null || inp === '')) {
              return {
                  type: 'required',
                  i18nLabel: i18n('isomorphic-schema--field_required', 'Required'),
                  message: "Obligatoriskt"
              };
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp) {
          return inp;
      }

  });

  var AnyOf = componentRegistry.createObjectPrototype({
      extends: [BaseField],
      implements: [IAnyOf],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);
          if (options) {
              this._valueTypes = options.valueTypes;
              delete options.valueTypes;
          }
      },

      // TODO: Validate against one of several
      validate: function validate(inp, options, context, async) {
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (inp) {
              if (async) {
                  var errorPromises = this._valueTypes.map(function (field) {
                      // Convert value so it can be checked properly, but if it converts to empty or undefined use original value
                      var tmp = field.fromString(inp);
                      return field.validateAsync(tmp || inp, options, context);
                  });

                  return Promise.all(errorPromises).then(function (fieldErrors) {
                      var error = fieldErrors.reduce(function (curr, next) {
                          // If one of the validators pass we are ok
                          if (next === undefined) {
                              return undefined;
                          } else {
                              return curr;
                          }
                      }, {
                          type: 'constraint_error',
                          i18nLabel: i18n('isomorphic-schema--any_of_error', 'The entered value doesn\'t match any of the allowed value types'),
                          message: 'Inmatat värde matchar inte de tillåtna alternativen'
                      });
                      return Promise.resolve(error);
                  });
              } else {
                  // Sync validate
                  var fieldErrors = this._valueTypes.map(function (field) {
                      // Convert value so it can be checked properly, but if it converts to empty or undefined use original value
                      var tmp = field.fromString(inp);
                      return field.validate(tmp || inp, options, context);
                  });
                  var error = fieldErrors.reduce(function (curr, next) {
                      // If one of the validators pass we are ok
                      if (next === undefined) {
                          return undefined;
                      } else {
                          return curr;
                      }
                  }, {
                      type: 'constraint_error',
                      i18nLabel: i18n('isomorphic-schema--any_of_error', 'The entered value doesn\'t match any of the allowed value types'),
                      message: 'Inmatat värde matchar inte de tillåtna alternativen'
                  });
                  return error;
              }
          } else {
              if (!async) {
                  return undefined;
              } else {
                  return Promise.resolve(undefined);
              }
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp) {
          return inp;
      }

  });

  var BoolField = componentRegistry.createObjectPrototype({
      implements: [IBoolField],

      constructor: function constructor(options) {},

      validateAsync: function validateAsync(inp, options, context) {
          return this.validate(inp, options, context, true);
      },

      validate: function validate(inp) {
          // False and undefined == false so this is always ok
          return undefined;
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp) {
          if (inp == "true" || inp === true) {
              return true;
          } else if (inp == "false" || inp === false) {
              return false;
          } else if (inp === "undefined" || inp === undefined) {
              return undefined;
          } else if (inp === 'null' || inp === null) {
              return null;
          } else {
              return inp;
          }
      }
  });

  var CreditCardField = componentRegistry.createObjectPrototype({
    implements: [ICreditCardField],

    extends: [BaseField],

    constructor: function constructor(options) {
      this._IBaseField.constructor(this, options);
      if (options) {
        this._cardValidationOptions = options.cardValidationOptions;
        delete options.cardValidationOptions;
      }
    },

    validate: function validate(inp) {
      var error = this._IBaseField.validate.call(this, inp);
      if (error) {
        return error;
      }

      if (inp) {
        var message;
        try {
          var result = _validateCreditCard(inp, this._cardValidationOptions);
        } catch (e) {
          var type = 'constraint_error';
          var i18nLabel = i18n('isomorphic-schema--credit_card_field_not_supported', 'Entered card type is not supported');
          message = "Inmatat kort stöds inte";
        } finally {
          if (!message && !(result.luhn_valid && result.length_valid)) {
            var type = 'type_error';
            var i18nLabel = i18n('isomorphic-schema--credit_card_field_incorrect_formatting', 'The card number is incorrectly entered');
            message = "Kortnumret ser inte korrekt ut";
          }
          if (message) {
            return {
              type: type,
              i18nLabel: i18nLabel,
              message: message
            };
          }
        }
      }
    },

    getMetaData: function getMetaData(inp) {
      return _validateCreditCard(inp);
    },

    toFormattedString: function toFormattedString(inp) {
      if (inp) {
        var pretty = function pretty(inp, type) {
          var formatter = _cardFormatters[type] || _cardFormatters['default'];
          return formatter(inp);
        };

        try {
          var res = this._validateCreditCard(inp, this._cardValidationOptions);
        } catch (e) {
          // 
        }

        return pretty(inp);
      } else {
        return "";
      }
    },

    fromString: function fromString(inp) {
      return inp.replace(/\s/g, '');
    }

  });

  // Helper method to validate credit cards
  var _validateCreditCard = function _validateCreditCard(inp, options) {
    var __indexOf = [].indexOf || function (item) {
      for (var i = 0, l = this.length; i < l; i++) {
        if (i in this && this[i] === item) return i;
      }return -1;
    };
    var card, card_type, card_types, get_card_type, is_valid_length, is_valid_luhn, normalize, validate, validate_number, _i, _len, _ref, _ref1;
    card_types = [{
      name: 'amex',
      pattern: /^3[47]/,
      valid_length: [15]
    }, {
      name: 'diners_club_carte_blanche',
      pattern: /^30[0-5]/,
      valid_length: [14]
    }, {
      name: 'diners_club_international',
      pattern: /^36/,
      valid_length: [14]
    }, {
      name: 'jcb',
      pattern: /^35(2[89]|[3-8][0-9])/,
      valid_length: [16]
    }, {
      name: 'laser',
      pattern: /^(6304|670[69]|6771)/,
      valid_length: [16, 17, 18, 19]
    }, {
      name: 'visa_electron',
      pattern: /^(4026|417500|4508|4844|491(3|7))/,
      valid_length: [16]
    }, {
      name: 'visa',
      pattern: /^4/,
      valid_length: [16]
    }, {
      name: 'mastercard',
      pattern: /^5[1-5]/,
      valid_length: [16]
    }, {
      name: 'maestro',
      pattern: /^(5018|5020|5038|6304|6759|676[1-3])/,
      valid_length: [12, 13, 14, 15, 16, 17, 18, 19]
    }, {
      name: 'discover',
      pattern: /^(6011|622(12[6-9]|1[3-9][0-9]|[2-8][0-9]{2}|9[0-1][0-9]|92[0-5]|64[4-9])|65)/,
      valid_length: [16]
    }];

    if (options == null) {
      options = {};
    }

    if ((_ref = options.accept) == null) {
      options.accept = function () {
        var _i, _len, _results;
        _results = [];
        for (_i = 0, _len = card_types.length; _i < _len; _i++) {
          card = card_types[_i];
          _results.push(card.name);
        }
        return _results;
      }();
    }
    _ref1 = options.accept;

    for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
      card_type = _ref1[_i];
      if (__indexOf.call(function () {
        var _j, _len1, _results;
        _results = [];
        for (_j = 0, _len1 = card_types.length; _j < _len1; _j++) {
          card = card_types[_j];
          _results.push(card.name);
        }
        return _results;
      }(), card_type) < 0) {
        throw "Credit card type '" + card_type + "' is not supported";
      }
    }

    get_card_type = function get_card_type(number) {
      var _j, _len1, _ref2;
      _ref2 = function () {
        var _k, _len1, _ref2, _results;
        _results = [];
        for (_k = 0, _len1 = card_types.length; _k < _len1; _k++) {
          card = card_types[_k];
          if (_ref2 = card.name, __indexOf.call(options.accept, _ref2) >= 0) {
            _results.push(card);
          }
        }
        return _results;
      }();
      for (_j = 0, _len1 = _ref2.length; _j < _len1; _j++) {
        card_type = _ref2[_j];
        if (number.match(card_type.pattern)) {
          return card_type;
        }
      }
      return null;
    };

    is_valid_luhn = function is_valid_luhn(number) {
      var digit, n, sum, _j, _len1, _ref2;
      sum = 0;
      _ref2 = number.split('').reverse();
      for (n = _j = 0, _len1 = _ref2.length; _j < _len1; n = ++_j) {
        digit = _ref2[n];
        digit = +digit;
        if (n % 2) {
          digit *= 2;
          if (digit < 10) {
            sum += digit;
          } else {
            sum += digit - 9;
          }
        } else {
          sum += digit;
        }
      }
      return sum % 10 === 0;
    };

    is_valid_length = function is_valid_length(number, card_type) {
      var _ref2;
      return _ref2 = number.length, __indexOf.call(card_type.valid_length, _ref2) >= 0;
    };

    validate_number = function validate_number(number) {
      var length_valid, luhn_valid;
      card_type = get_card_type(number);
      luhn_valid = false;
      length_valid = false;
      if (card_type != null) {
        luhn_valid = is_valid_luhn(number);
        length_valid = is_valid_length(number, card_type);
      }
      return {
        card_type: card_type,
        luhn_valid: luhn_valid,
        length_valid: length_valid
      };
    };

    validate = function validate() {
      var number;
      number = normalize(inp);
      return validate_number(number);
    };

    normalize = function normalize(number) {
      return number.replace(/[ -]/g, '');
    };

    return validate();
  };

  var _cardFormatters = {
    default: function _default(inp) {
      var tmp = inp.match(/.{1,4}/g);
      return tmp.join(" ");
    }
  };

  var TextField = componentRegistry.createObjectPrototype({
      implements: [ITextField],
      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);

          if (options) {
              this._minLength = options.minLength;
              delete options.minLength;
              this._maxLength = options.maxLength;
              delete options.maxLength;
              this._trim = options.trim;
              delete options.trim;
          }
      },

      validate: function validate(inp) {
          if (this._trim) {
              inp = inp.trim();
          }

          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (!this._isRequired && (inp === null || typeof inp === "undefined" || inp === '')) {
              return;
          }

          // TODO: Check for line breaks
          if (inp && typeof inp !== "string") {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--text_field_no_string', 'The field doesn\'t contain text'),
                  message: "Fältet är inte en sträng"
              };
          } else if (this._minLength && inp.length < this._minLength) {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--text_field_too_short', 'The text is too short. Min ${minLength} chars.'),
                  message: "Texten är för kort. Minst " + this._minLength + " tkn"
              };
          } else if (this._maxLength && inp.length > this._maxLength) {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--text_field_too_long', 'The text is too long. Max ${maxLength} chars.'),
                  message: "Texten är för lång. Max " + this._maxLength + " tkn"
              };
          }
      },

      fromString: function fromString(inp) {
          if (this._trim) {
              return inp.trim();
          } else {
              return inp;
          }
      }
  });

  var isValid = require('date-fns/is_valid');

  var DateField = componentRegistry.createObjectPrototype({
      implements: [IDateField],

      extends: [TextField],

      constructor: function constructor(options) {
          this._ITextField.constructor.call(this, options);
      },

      validate: function validate(inp) {
          var error = this._ITextField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (inp && (inp.length != 10 || !isValid(new Date(inp)))) {
              error = {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--date_field_incorrect_formatting', 'This doesn\'t look like a date'),
                  message: "Det ser inte ut som datum"
              };

              return error;
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp) {
          return inp;
      }
  });

  var parse = require('date-fns/parse');
  var format = require('date-fns/format');

  var DateTimeField = componentRegistry.createObjectPrototype({
      implements: [IDateTimeField],

      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);
      },

      validate: function validate(inp) {
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (inp && !(inp instanceof Date)) {
              error = {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--date_time_field_incorrect_formatting', 'This doesn\'t look like a date with time'),
                  message: "Det ser inte ut som ett riktigt datumobjekt med tid"
              };

              return error;
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return format(inp, 'YYYY-MM-DDTHH:mm:ssZ');
      },

      fromString: function fromString(inp) {
          return parse(inp);
      }
  });

  var reDecimal = /[^0-9\.]/g;

  var DecimalField = componentRegistry.createObjectPrototype({
      implements: [IDecimalField],

      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);
          if (options) {
              this._minValue = options.min;
              delete options.min;
              this._maxValue = options.max;
              delete options.max;
              this._precision = options.precision && options.precision >= 0 && options.precision;
              delete options.precision;
          }
      },

      validate: function validate(inp) {
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (!this._isRequired && (inp === null || typeof inp === "undefined" || inp === '')) {
              return;
          }

          if (typeof inp !== "number" || isNaN(inp)) {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--decimal_field_not_number', 'The field doesn\'t contain numbers'),
                  message: "Värdet innehåller annat än siffror"
              };
          }

          if (typeof this._minValue !== 'undefined' && parseFloat(inp) < this._minValue) {
              return {
                  type: 'constraint_error',
                  i18nLabel: i18n('isomorphic-schema--decimal_field_too_small', 'The value is too small. Min ${minValue}'),
                  message: "Minimum " + this._minValue
              };
          }

          if (typeof this._maxValue !== 'undefined' && parseFloat(inp) > this._maxValue) {
              return {
                  type: 'constraint_error',
                  i18nLabel: i18n('isomorphic-schema--decimal_field_too_big', 'The value is too big. Max ${maxValue}'),
                  message: "Max " + this._maxValue
              };
          }
      },

      toFormattedString: function toFormattedString(inp) {
          var outp = inp + '';

          if (inp === null || typeof inp === "undefined") {
              return inp;
          }
          // Print only nrof decimals shown in precision if set
          if (typeof this._precision !== 'undefined') {
              var tmp = outp.split('.');
              outp = tmp[0];
              if (this._precision > 0) {
                  outp += '.';
                  if (tmp.length > 1) {
                      outp += (tmp[1] + '').substr(0, this._precision);
                  }
              }
          }
          return outp;
      },

      fromString: function fromString(inp) {
          if (typeof inp === "string") {
              var tmp = parseFloat(inp.replace(reDecimal, ""));
          } else {
              var tmp = parseFloat(inp);
          }
          if (isNaN(tmp)) {
              return undefined;
          } else {
              // Round to precision if set
              if (typeof this._precision !== 'undefined') {
                  var tmp = tmp * Math.pow(10, this._precision);
                  var tmp = Math.round(tmp);
                  var tmp = tmp / Math.pow(10, this._precision);
              }

              return tmp;
          }
      }
  });

  var DynamicSelectBaseField = componentRegistry.createObjectPrototype({
      implements: [IDynamicSelectBaseField],

      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);
      },

      validate: function validate(inp, options, context) {
          // Check that this isn't undefined if it is required
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          var error = inp && this.valueType.validate(inp);
          if (error) {
              return error;
          }

          // Since we have passed the required test we can just check if the value is undefined
          // or null and return field errors undefined 
          if (inp === null || typeof inp === 'undefined') {
              return;
          }

          var options = this.getOptions(inp, options, context);

          var matches = false;
          for (var i = 0; i < options.length; i++) {
              if (options[i].name === inp) {
                  matches = true;
                  break;
              }
          }

          if (!matches) {
              error = {
                  type: 'constraint_error',
                  i18nLabel: i18n('isomorphic-schema--select_field_value_error', 'The selected value is not allowed'),
                  message: "Valt värde finns inte i listan över tillåtna värden"
                  //console.log(error)
              };return error;
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return this.valueType.fromString(inp);
      },

      fromString: function fromString(inp) {
          return this.valueType.fromString(inp);
      },

      getOptions: function getOptions(inp, options, context) {},

      getOptionTitle: function getOptionTitle(inp, options, context) {
          var selectOptions = this.getOptions(inp, options, context);
          for (var i = 0; i < selectOptions.length; i++) {
              if (selectOptions[i].name === inp) {
                  return selectOptions[i].title;
              }
          }
      }
  });

  var DynamicSelectAsyncBaseField = componentRegistry.createObjectPrototype({
      implements: [IDynamicSelectAsyncBaseField],

      extends: [DynamicSelectBaseField],

      constructor: function constructor(options) {
          this._IDynamicSelectBaseField.constructor.call(this, options);
      },

      // DynamicSelectAsyncBaseField:
      validateAsync: function validateAsync(inp, options, context) {
          // Check that this isn't undefined if it is required
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return Promise.resolve(error);
          }
          // Since we have passed the required test we can just check if the value is undefined
          // or null and return field errors undefined 
          if (inp === null || typeof inp === 'undefined') {
              return Promise.resolve(undefined);
          }

          if (inp) {
              return this.valueType.validateAsync(inp, options, context);
          } else {
              return Promise.resolve(undefined);
          }

          /*
           Example implementation:
           var result = this._IDynamicSelectAsyncBaseField.validateAsync.call(this, inp)
           // Check if we failed validation in DynamicSelectAsyncBaseField
          if (result) {
              return result
          }
           return doAsyncCall(inp)
              .then(function (didMatch) {
                  if (didMatch) {
                      return Promise.resolve(undefined)
                  } else {
                      return Promise.resolve({
                          type: 'constraint_error',
                          i18nLabel: i18n('isomorphic-schema--select_field_value_error', 'The selected value is not allowed'),
                          message: "Valt värde finns inte i listan över tillåtna värden"
                      })
                  }
              })
              .catch(function (err) {
                  throw new Error('Couldn\t perform validation: ' + err.message)
              })
          */
      },

      validate: function validate() {
          throw new Error("You tried to call a field extending DynamicSelectAsyncBaseField synchronously, needs to be async. Use .validateAsync(...) instead!");
      },

      // This is used by formlibs
      getOptionsAsync: function getOptionsAsync(inp, options, context) {
          /*
          return doAsyncCall(inp)
              .then(function (results) {
                  var outp = results.map(function (item) {
                      return { name: item.id, title: item.title }
                  })
                  return Promise.resolve(outp)
              })
              .catch(function (err) {
                  throw new Error('Couldn\t fetch options: ' + err.message)
              })
          */
      },

      getOptions: function getOptions() {
          throw new Error("You tried to call a field extending DynamicSelectAsyncBaseField synchronously, needs to be async. Use .getOptionsAsync(...) instead!");
      },

      // This is used by formlibs
      getOptionTitleAsync: function getOptionTitleAsync(inp, options, context) {
          /*
          return doAsyncCall(inp)
              .then(function (title) {
                  if (title) {
                      return Promise.resolve(title)
                  } else {
                      return Promise.resolve({
                          type: 'value_error',
                          i18nLabel: i18n('isomorphic-schema--select_field_value_error', 'The selected value is not allowed'),
                          message: "Valt värde finns inte i listan över tillåtna värden"
                      })
                  }
              })
              .catch(function (err) {
                  throw new Error('Couldn\t fetch title: ' + err.message)
              })
          */
      },

      getOptionTitle: function getOptionTitle() {
          throw new Error("You tried to call a field extending DynamicSelectAsyncBaseField synchronously, needs to be async. Use .getOptionTitleAsync(...) instead!");
      },

      toFormattedString: function toFormattedString(inp) {
          return this.valueType.fromString(inp);
      },

      fromString: function fromString(inp) {
          return this.valueType.fromString(inp);
      }
  });

  var EmailField = componentRegistry.createObjectPrototype({
      implements: [IEmailField],

      extends: [TextField],

      constructor: function constructor(options) {
          this._ITextField.constructor.call(this, options);
      },

      validate: function validate(inp) {
          var error = this._ITextField.validate.call(this, inp);
          if (error) {
              return error;
          }

          // Required has been checked so if it is empty it is ok
          if (!inp) {
              return;
          }

          /* 
              Using a proper e-mail testing algorithm
              http://thedailywtf.com/articles/Validating_Email_Addresses
          */
          var add = inp;
          var ampisthere = false;
          var spacesthere = false;

          var textbeforeamp = false;
          var textafteramp = false;
          var dotafteramp = false;
          var othererror = false;

          for (var i = 0; i < add.length; ++i) {
              if (add.charAt(i) == '@') {
                  if (ampisthere) othererror = true;

                  ampisthere = true;
              } else if (!ampisthere) textbeforeamp = true;else if (add.charAt(i) == '.') dotafteramp = true;else textafteramp = true;

              if (add.charAt(i) == ' ' || add.charAt(i) == ',') spacesthere = true;
          }

          if (spacesthere || !ampisthere || !textafteramp || !textbeforeamp || !dotafteramp || othererror) {
              error = {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--email_field_incorrect_formatting', 'This is not a valid e-mail address'),
                  message: "Det här är inte en riktig e-postadress"
              };

              return error;
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp) {
          return inp;
      }
  });

  var TextAreaField = componentRegistry.createObjectPrototype({
      implements: [ITextAreaField],
      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);

          if (options) {
              this._minLength = options.minLength;
              delete options.minLength;
              this._maxLength = options.maxLength;
              delete options.maxLength;
          }
      },

      validate: function validate(inp) {
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (!this._isRequired && (inp === null || typeof inp === "undefined" || inp === '')) {
              return;
          }

          if (inp && typeof inp !== "string") {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--text_area_field_no_string', 'The field doesn\'t contain text'),
                  message: "Fältet är inte en sträng"
              };
          } else if (this._minLength && inp.length < this._minLength) {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--text_field_too_short', 'The text is too short. Min ${minLength} chars.'),
                  message: "Texten är för kort. Minst " + this._minLength + " tkn"
              };
          } else if (this._maxLength && inp.length > this._maxLength) {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--text_field_too_long', 'The text is too long. Max ${maxLength} chars.'),
                  message: "Texten är för lång. Max " + this._maxLength + " tkn"
              };
          }
      }
  });

  var striptags = require('striptags');

  var HTMLAreaField = componentRegistry.createObjectPrototype({
      implements: [IHTMLAreaField],
      extends: [TextAreaField],

      constructor: function constructor(options) {
          this._ITextAreaField.constructor.call(this, options);
      },

      validate: function validate(inp) {
          var tmp = inp;
          if (typeof inp === 'string') {
              tmp = striptags(inp);
          }
          var error = this._ITextAreaField.validate.call(this, tmp);
          if (error) {
              return error;
          }

          // TODO: should I add HTML validation? Might need cheerio, in which case 
          // perhaps only serverside validation to keep file size small in browser?
          // Could use striptags with allowed tags but it doesn't support attribute
          // stripping
      }
  });

  var reInteger = /[^0-9\.]/g;

  var IntegerField = componentRegistry.createObjectPrototype({
      implements: [IIntegerField],

      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);
          if (options) {
              this._minValue = options.min;
              delete options.min;
              this._maxValue = options.max;
              delete options.max;
          }
      },

      validate: function validate(inp) {
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (!this._isRequired && (inp === null || typeof inp === "undefined" || inp === '')) {
              return;
          }

          if (typeof inp !== "number" || isNaN(inp)) {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--integer_field_not_number', 'The field doesn\'t contain numbers'),
                  message: "Värdet innehåller annat än siffror"
              };
          }

          if (parseInt(inp) !== inp) {
              return {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--integer_field_no_decimals', 'The field may not contain decimals'),
                  message: "Värdet får inte innehålla en decimaldel"
              };
          }

          if (typeof this._minValue !== 'undefined' && inp < this._minValue) {
              return {
                  type: 'constraint_error',
                  i18nLabel: i18n('isomorphic-schema--integer_field_too_small', 'The value is too small. Min ${minValue}'),
                  message: "Minimum " + this._minValue
              };
          }

          if (typeof this._maxValue !== 'undefined' && inp > this._maxValue) {
              return {
                  type: 'constraint_error',
                  i18nLabel: i18n('isomorphic-schema--integer_field_too_big', 'The value is too big. Max ${maxValue}'),
                  message: "Max " + this._maxValue
              };
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return typeof inp === 'undefined' || inp === null ? '' : inp + '';
      },

      fromString: function fromString(inp) {
          if (typeof inp === "string") {
              var tmp = parseInt(inp.replace(reInteger, ""));
          } else {
              var tmp = parseInt(inp);
          }
          if (isNaN(tmp)) {
              return undefined;
          } else {
              return tmp;
          }
      }
  });

  var ListField = componentRegistry.createObjectPrototype({
      implements: [IListField],

      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);

          if (options) {
              // Always set valueType to required to get validation per item
              // Should be an instance of a field.
              this.valueType = options.valueType || options.valueFieldType;
              this.options = options.options;
              this._minItems = options.minItems;
              delete options.minItems;
              this._maxItems = options.maxItems;
              delete options.maxItems;
          }
      },

      validate: function validate(inp, options, context, async) {
          if (inp && inp.length == 0) {
              inp = undefined;
          }
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          // If it is undefined or null, then we just return 
          // that this value is ok
          if (!inp) {
              return undefined;
          }

          // Check that this is an array
          if (!(inp instanceof Array)) {
              error = {
                  type: 'type_error',
                  i18nLabel: i18n('isomorphic-schema--list_field_type_error', 'This is not a proper list. This is a bug in the application'),
                  message: "This is not a proper list. This is a bug in the application"
                  //console.log(error)
              };return error;
          }

          if (this._maxItems && inp.length > this._maxItems) {
              error = {
                  type: 'value_error',
                  i18nLabel: i18n('isomorphic-schema--list_field_value_error_too_many_items', 'Too many items in list, max ${maxItems} allowed'),
                  message: 'Too many items in list, max ${maxItems} allowed'
              };
          }

          if (this._minItems && inp.length < this._minItems) {
              error = {
                  type: 'value_error',
                  i18nLabel: i18n('isomorphic-schema--list_field_value_error_too_few_items', 'Too few items in list, min ${minItems} allowed'),
                  message: 'Too few items in list, min ${minItems} allowed'
              };
          }

          var tmpValidationErrors = [];
          var validationPromises = [];
          inp.forEach(function (item, i) {
              // Make sure we have a new options object...
              if (!options) {
                  var newOptions = {};
              } else {
                  var newOptions = cloneShallow(options);
              }

              // ...so we can add objectPath to determine where we are
              if (Array.isArray(newOptions.objectPath)) {
                  newOptions.objectPath = newOptions.objectPath.concat([i]);
              } else {
                  newOptions.objectPath = [i];
              }

              if (!async) {
                  var tmpError = this.valueType.validate(item, newOptions, context);
                  if (tmpError) {
                      tmpValidationErrors.push({ id: i, error: tmpError });
                  }
              } else {
                  var promise = this.valueType.validateAsync(item, newOptions, context);
                  // Do not return this promise, add it to validationPromises. We will 
                  // handle the list of validationPromises further down
                  if (promise && promise.then) {
                      validationPromises.push(promise.then(function (tmpError) {
                          if (tmpError) {
                              return Promise.resolve({ id: i, error: tmpError });
                          }
                      }));
                  }
              }
          }.bind(this));

          if (!async || validationPromises.length === 0) {
              if (tmpValidationErrors.length > 0) {
                  var tmpErrors = {};
                  tmpValidationErrors.map(function (item) {
                      tmpErrors[item.id] = item.error;
                  });

                  if (error !== undefined) {
                      error.errors = tmpErrors;
                  } else {
                      error = {
                          type: 'list_error',
                          i18nLabel: i18n('isomorphic-schema--list_field_value_error', 'There is an error in the content of this list'),
                          message: "There is an error in the content of this list",
                          errors: tmpErrors
                      };
                  }
              }
              // If there weren't any errors this will return undefined
              return error;
          } else {
              return Promise.all(validationPromises).then(function (errors) {
                  // Filter out non-errors (undefined)
                  errors = errors.filter(function (error) {
                      return error;
                  });
                  if (errors.length === 0) {
                      return Promise.resolve(undefined);
                  } else {
                      var tmpErrors = {};
                      errors.forEach(function (err) {
                          tmpErrors[err.id] = err.error;
                      });
                      return Promise.resolve({
                          type: 'list_error',
                          i18nLabel: i18n('isomorphic-schema--list_field_value_error', 'There is an error in the content of this list'),
                          message: "There is an error in the content of this list",
                          errors: tmpErrors
                      });
                  }
              });
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp) {
          var _this = this;
          if (inp instanceof Array) {
              return inp.map(function (item) {
                  return _this.valueType.fromString(item);
              });
          } else {
              return inp;
          }
      }
  });

  var MultiSelectField = componentRegistry.createObjectPrototype({
      implements: [IMultiSelectField],

      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);

          if (options) {
              // Always set valueType to required to get validation per item
              // Should be an instance of a field.
              this.valueType = options.valueType;
              this.options = options.options;
          }
      },

      validate: function validate(inp, options, context, async) {
          // Check that this isn't undefined if it is required
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          // If undefined and not required just return ok
          if (typeof inp === 'undefined' || inp === null) {
              return;
          }

          // We need to check every item in the list to see that they are valid
          var errors = inp.map(function (item) {
              // Chack value is of valueType
              var error = this.valueType.validate(item);
              if (typeof error !== 'undefined') {
                  if (!aync) {
                      return error;
                  } else {
                      return Promise.resolve(error);
                  }
              }

              var options = this.options;

              // Check that selected value is in options list
              var matches = false;
              for (var i = 0; i < options.length; i++) {
                  if (options[i].name === item) {
                      matches = true;
                      break;
                  }
              }
              if (!matches) {
                  return {
                      type: 'constraint_error',
                      i18nLabel: i18n('isomorphic-schema--multi_select_field_value_error', 'One or more of the selected values is not allowed'),
                      message: "Ett eller flera valda värden finns inte i listan över tillåtna värden"
                  };
              }
          }.bind(this)).filter(function (item) {
              return typeof item !== 'undefined';
          });
          if (!async || errors.length === 0) {
              return errors[0];
          } else {
              return Promise.all(errors).then(function (results) {
                  var results = results.filter(function (item) {
                  });
                  return Promise.resolve(results[0]);
              });
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp) {
          return inp;
      },

      getOptionTitle: function getOptionTitle(name) {
          for (var i = 0; i < this.options.length; i++) {
              if (this.options[i].name === name) return this.options[i].title;
          }
      }
  });

  /*
      Object field
  */

  var ObjectField = componentRegistry.createObjectPrototype({
      implements: [IObjectField],
      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);
          if (options) {
              this._schema = options.schema;
              this._interface = options.interface;
              this.objectFactoryName = options.objectFactoryName;
          }
      },

      validate: function validate(inp, options, context, async) {
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          // Validate data and return an error object
          if (inp) {

              if (this._interface) {
                  if (!async) {
                      var schema = this._interface.schema;
                  }
              } else {
                  var schema = this._schema;
              }

              // Allow object fields without schema (useful for modelling objects that we don't know what they look like)
              if (typeof schema === 'undefined') {
                  return async ? Promise.resolve(undefined) : undefined;
              }

              if (!async) {
                  var formErrors = schema.validate(inp, options, context, async);
                  if (formErrors) {
                      // report error if data is passed and errors are found
                      return {
                          type: "object_error",
                          i18nLabel: i18n('isomorphic-schema--object_field_value_error', 'There is an error in the content of this object'),
                          message: "Delformuläret innehåller fältfel",
                          fieldErrors: formErrors.fieldErrors,
                          invariantErrors: formErrors.invariantErrors
                      };
                  }
              } else {
                  var promise = schema.validateAsync(inp, options, context);

                  if (!(promise && promise.then)) {
                      return Promise.resolve(promise);
                  } else {
                      return promise.then(function (formErrors) {
                          return Promise.resolve(formErrors);
                      });
                  }
              }
          } else {
              if (!async) {
                  return undefined;
              } else {
                  return Promise.resolve(undefined);
              }
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return inp;
      },

      fromString: function fromString(inp, options) {
          var schema = this._interface ? this._interface.schema : this._schema;

          if ((typeof inp === 'undefined' ? 'undefined' : _typeof(inp)) === 'object' && typeof schema !== 'undefined') {
              return schema.transform(inp, options);
          } else {
              return inp;
          }
      }

  });

  /*
      Object field
  */

  var ObjectRelationField = componentRegistry.createObjectPrototype({
      implements: [IObjectRelationField],
      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);
          if (options) {
              this._resolverName = options.resolverName;
              this._interface = options.interface;
          }
      },

      validate: function validate(inp, options) {
          // We need to resolve the relation to validate it
          var val = inp.get();

          var error = this._IBaseField.validate.call(this, val);
          if (error) {
              return error;
          }

          /*
          // TODO: Implement object type validation
          // TODO: Consider implementing test on server for existence of related object
          // Validate data and return an error object
          if (val) {
              if (this._inte && this._objectType !== val._type) {
                  
                  // The related object doesn't implement any of the accepted interfaces
                  return {
                      type: "relation_error",
                      message: "Relaterat objekt är av fel typ, ska vara: " + this._objectType
                  }
              }            
          } else {
              return undefined
          }
          */
          return undefined;
      }

  });

  var OrgNrField = componentRegistry.createObjectPrototype({
      implements: [IOrgNrField],

      extends: [TextField],

      constructor: function constructor(options) {
          this._ITextField.constructor.call(this, options);
      },

      validate: function validate(inp) {
          var error = this._ITextField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (inp) {
              var message;
              try {
                  var error = _validateOrgNr(inp);
              } catch (e) {
                  var type = 'value_error';
              } finally {
                  if (!message && error) {
                      var type = 'type_error';
                      message = error.message;
                      var i18nLabel = error.i18nLabel;
                  }
                  if (message) {
                      return {
                          type: type,
                          i18nLabel: i18nLabel,
                          message: message
                      };
                  }
              }
          }
      },

      toFormattedString: function toFormattedString(inp) {
          if (inp) {
              var tmp = inp.match(/.{1,8}/g);
              var outp = tmp.join('-');
              return outp.length === 8 ? outp + '-' : outp;
          } else {
              return "";
          }
      },

      fromString: function fromString(inp) {
          var tmp = inp.replace(/([^0-9]*)/g, '');
          if (tmp.length > 12) {
              tmp = tmp.split("");
              tmp = tmp.splice(0, 12);
              tmp = tmp.join("");
          }
          return tmp;
      }
  });

  var _validateOrgNr = function _validateOrgNr(inp, options) {
      /*
             8  1 1 2 1 8  9 8  7 6
          *  2  1 2 1 2 1  2 1  2 1
          -------------------------
             ^  ^ ^ ^ ^ ^  ^ ^  ^ 
            16  1 2 2 2 8 18 8 14 6
      
          1+6+1+2+2+2+8+1+8+8+1+4+6 = 50
      */

      var inpLst = inp.split("");
      if (inpLst.length < 12) {
          return {
              i18nLabel: i18n('isomorphic-schema--org_nr_field_too_short', 'Entered number is too short'),
              message: "Inmatat nummer är för kort"
          };
      }

      var luhLst = [0, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1];
      var calcLst = luhLst.map(function (val, i) {
          return luhLst[i] * parseInt(inpLst[i]);
      });
      var tmp = 0;
      calcLst.map(function (val) {
          tmp += val % 10 + Math.floor(val / 10);
      });
      if (tmp % 10 == 0) {
          return undefined;
      } else {
          return {
              i18nLabel: i18n('isomorphic-schema--org_nr_field_wrong_checksum', 'The entered number is incorrect (checksum error)'),
              message: "Kontrollsiffran stämmer inte"
          };
      }
  };

  var PasswordField = componentRegistry.createObjectPrototype({
      implements: [IPasswordField],

      extends: [TextField],

      constructor: function constructor(options) {
          this._ITextField.constructor.call(this, options);
      },

      validate: function validate(inp) {
          var error = this._ITextField.validate.call(this, inp);
          if (error) {
              return error;
          }

          if (inp !== null && typeof inp !== 'undefined' && inp.length < 8) {
              error = {
                  type: 'constraint_error',
                  i18nLabel: i18n('isomorphic-schema--password_field_too_short', 'The password must contain at least 8 chars'),
                  message: "Lösenordet måste innehålla minst 8 tecken"
              };

              return error;
          }
      }
  });

  var SelectField = componentRegistry.createObjectPrototype({
      implements: [ISelectField],

      extends: [BaseField],

      constructor: function constructor(options) {
          this._IBaseField.constructor.call(this, options);

          if (options) {
              // Always set valueType to required to get validation per item
              // Should be an instance of a field.
              this.valueType = options.valueType;
              this.options = options.options;
          }
      },

      validate: function validate(inp, options, context, async) {
          var error = this._IBaseField.validate.call(this, inp);
          if (error) {
              return error;
          }

          var error = inp && this.valueType.validate(inp);
          if (error) {
              return error;
          }

          //console.log("[Select] options:")
          //console.log(this.options)

          //console.log("[Select] current:")
          //console.log(inp)

          // Since we have passed the required test we can just check if the value is undefined
          // or null and return field errors undefined 
          if (inp === null || typeof inp === 'undefined') {
              return;
          }

          var options = this.options;

          var matches = false;
          for (var i = 0; i < options.length; i++) {
              if (options[i].name === inp) {
                  matches = true;
                  break;
              }
          }

          if (!matches) {
              error = {
                  type: 'constraint_error',
                  i18nLabel: i18n('isomorphic-schema--select_field_value_error', 'The selected value is not allowed'),
                  message: "Valt värde finns inte i listan över tillåtna värden"
                  //console.log(error)
              };return error;
          }
      },

      toFormattedString: function toFormattedString(inp) {
          return this.valueType.fromString(inp);
      },

      fromString: function fromString(inp) {
          return this.valueType.fromString(inp);
      },

      getOptionTitle: function getOptionTitle(inp) {
          for (var i = 0; i < this.options.length; i++) {
              if (this.options[i].name === inp) return this.options[i].title;
          }
      }
  });

  var allowedCharsRegex = /[^\d().\/+ -]+/;
  var TelephoneField = componentRegistry.createObjectPrototype({
      implements: [ITelephoneField],

      extends: [TextField],

      constructor: function constructor(options) {
          this._ITextField.constructor.call(this, options);
      },

      validate: function validate(inp) {
          var error = this._ITextField.validate.call(this, inp);
          if (error) {
              return error;
          }

          // Required has been checked so if it is empty it is ok
          if (!inp) {
              return;
          }

          var tmpInp = inp.replace(/[^\d]/, '');
          if (tmpInp.length > 15) {
              return {
                  type: 'value_error',
                  i18nLabel: i18n('isomorphic-schema--telephone_field_too_long', 'This is not a valid telephone number, max 15 numericals'),
                  message: "This is not a valid telephone number, max 15 numericals"
              };
          }

          if (allowedCharsRegex.test(inp)) {
              return {
                  type: 'value_error',
                  i18nLabel: i18n('isomorphic-schema--telephone_field_invalid_chars', 'Invalid telephone number, you may only use numbers and common delimiters'),
                  message: "Invalid telephone number, you may only use numbers and common delimiters"
              };
          }
      }
  });

  exports.interfaces = interfaces;
  exports.Schema = Schema;
  exports.i18n = i18n;
  exports.renderString = renderString;
  exports.AnyOf = AnyOf;
  exports.BaseField = BaseField;
  exports.BoolField = BoolField;
  exports.CreditCardField = CreditCardField;
  exports.DateField = DateField;
  exports.DateTimeField = DateTimeField;
  exports.DecimalField = DecimalField;
  exports.DynamicSelectAsyncBaseField = DynamicSelectAsyncBaseField;
  exports.DynamicSelectBaseField = DynamicSelectBaseField;
  exports.EmailField = EmailField;
  exports.HTMLAreaField = HTMLAreaField;
  exports.IntegerField = IntegerField;
  exports.ListField = ListField;
  exports.MultiSelectField = MultiSelectField;
  exports.ObjectField = ObjectField;
  exports.ObjectRelationField = ObjectRelationField;
  exports.OrgNrField = OrgNrField;
  exports.PasswordField = PasswordField;
  exports.SelectField = SelectField;
  exports.TelephoneField = TelephoneField;
  exports.TextAreaField = TextAreaField;
  exports.TextField = TextField;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=index.js.map
